/*
 * CLDAGibbs.h
 *
 *  Created on: May 21, 2010
 *      Author: pdas
 */

#ifndef CLDAGIBBS_H_
#define CLDAGIBBS_H_

#include "../DataSet/Dataset.h"
#include "AGraphicalModel.h"
#include "CLDAGeneralModel.h"
#include "../SamplingAlgorithms/CLDASamplingFunctionoid.h"
#include "../include/CUtilities.h"
#include "../include/TemplateMath.h"

/**
 * @author Pradipto Das. email: pdas3@buffalo.edu
 *
 * @brief
 * Class CLDAGibbs - implements Gibbs sampling for the Latent Dirichlet Allocation model
 *
 * This class packages up
 * - the model data structures
 * - the functions required to create, initialize, load and save those data structures
 * - friend classes whenever applicable
 *
 * @see base class CLDAGeneralModel
 * @see psiexp.ss.uci.edu/research/papers/sciencetopics.pdf
 */
class CLDAGibbs : public CLDAGeneralModel {

protected:
	/*!
	 * @brief class CLDASamplingFunctionoid is a friend of this class
	 *
	 * In a Gibbs sampling scenario, the functionoid invokes the imputation() and posterior() steps
	 * that continuously manipulate the model parameter values until the desired number of iterations
	 * or the termination condition
	 *
	 * In an EM or VBEM scenario, the functionoid might invoke the expectation() and maximization() steps
	 * or the Veriational-Bayesian expectation() and Variational-Bayesian maximization() steps
	 *
	 * @note
	 * None of the the EM or VBEM framework is available in this tolbox yet, but can be easily implemented
	 */
	friend class CLDASamplingFunctionoid;
	/*!
	 * @brief This is needed for printing topics as datum labels to file or console
	 */
	std::string _datum_to_id_dict_filename;
	/*!
	 * @brief This member variable contains the absolute filename of the saved topic datum multinomial parameters
	 */
	std::string _model_train_saved_topic_mult_filename;
	/*!
	 * @brief This member variable indicates whether to read data in grouped format or sequential format.
	 * Default value is the std::string "grouped".
	 * Only two values are allowed - "grouped" or "sequential"
	 */
	std::string _data_input_mode;
	/*!
	 * @brief a boolean to accomodate for unknown word ids
	 */
	bool _accommodate_for_unknown_datum;
	/*!
	 * @brief the total datum count in the training corpus - useful for measuring Bayesian Information Criterion
	 */
	long _training_datum_total_count;

	/////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////// data structures for TRAINING
	/////////////////////////////////////////////////////////////////////////////////////////////////
protected:
	/**
	 * \defgroup MEMBER_VARIABLES_FOR_TRAINING variables for training for LDAGibbs
	 */
	/*@{*/

	//! The pointer to SCorpus that represents the training corpus
	SCorpus* _corpus;

	//! @brief ( symmetric ) dirichlet prior for beta (topics themselves)
	double _eta;
	//! @brief ( symmetric ) dirichlet prior for theta (topic proportions)
	double _alpha;
	//! @brief A D x K matrix (document x topic)
	std::vector<std::vector<double>* >* _theta;
	//! @brief A K x V matrix (topic x vocabulary)
	std::vector<std::vector<double>* >* _beta;
	//! @brief number of times topic k is assigned to document d
	std::vector<std::vector<int>* >* _z_d_k;
	//! @brief number of times topic k is assigned to a word v
	std::vector<std::vector<int>* >* _z_v_k;
	//! @brief row-wise total of _z_d_k    - how many times topic k appears in document d
	std::vector<int>*		   _z_d_k_total;
	//! @brief column-wise total of _z_v_k - how many times word v appears in topic k
	std::vector<int>*		   _z_v_k_total;
	//! @brief A D x N_d matrix
	std::vector<std::vector<int>* >* _z_d_n;

	//! @brief full path for the model index file
	std::string _model_index_filename;
	//! @brief full path of the directory to store the output files
	std::string _model_root_dir;
	//! @brief number of topics
	size_t _K;
	//! @brief number of datum in the vocabulary
	size_t _V;
	//! @brief number of docs in the corpus
	size_t _D;

	//! @brief number of gibbs iterations
	size_t _n_gibbs_iter;
	//! @brief number of iterations before computing posteriors
	size_t _gibbs_iter_lag;
	//! @brief number of iterations for sampler burning in
	size_t _gibbs_iter_burnin;

	//! @brief likelihood under this model for a given value of K
	double _log_prob;
	/*@}*/

	/////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////// data structures for TESTING
	/////////////////////////////////////////////////////////////////////////////////////////////////
protected:
	/**
	 * \defgroup MEMBER_VARIABLES_FOR_TESTING variables for testing or validating for LDAGibbs
	 */
	/*@{*/
	//! The pointer to SCorpus that represents the test or validation corpus
	SCorpus* _corpus_test;

	//double _eta;								// ( symmetric ) dirichlet prior for beta (topics themselves) - remains the same as in training
	//double _alpha;							// ( symmetric ) dirichlet prior for theta (topic proportions) - remains the same as in training
	//! A D_test x K matrix
	std::vector<std::vector<double>* >* _theta_test;
	//! A K x V_test matrix
	std::vector<std::vector<double>* >* _beta_test;
	//! number of times topic k is assigned to test document d
	std::vector<std::vector<int>* >* _z_d_k_test;
	//! number of times topic k is assigned to a test word v
	std::vector<std::vector<int>* >* _z_v_k_test;
	//! row-wise total of _z_d_k    - how many times topic k appears in test document d
	std::vector<int>*		   _z_d_k_total_test;
	//! column-wise total of _z_v_k - how many times test word v appears in topic k
	std::vector<int>*		   _z_v_k_total_test;
	//! A D_test x N_d matrix
	std::vector<std::vector<int>* >* _z_d_n_test;
	//! number of times topic k is assigned to a training word v
	std::vector<std::vector<int>* >* _z_v_k_full;
	//! column-wise total of _z_v_k - how many times training word v appears in topic k
	std::vector<int>*		   _z_v_k_total_full;

	//! full path for the model index file for testing
	std::string _model_index_filename_test;
	//! full path of the directory of stored the training files
	std::string _model_train_root_dir;
	//std::string _model_root_dir;					// full path of the directory to store the output files - remains the same as in training
	//size_t _K;								// number of topics - remains the same as in training
	//! number of terms in the test vocabulary
	size_t _V_test;
	//! number of docs in the test corpus
	size_t _D_test;

	//size_t _n_gibbs_iter_test;				// number of gibbs iterations
	//size_t _gibbs_iter_lag_test;				// number of iterations before computing posteriors
	//size_t _gibbs_iter_burnin_test;			// number of iterations for sampler burning in

	//double _log_prob_test;					// likelihood under this model for a given value of K

	//map<size_t, size_t> _test_datum_id_to_sequential_index_id_map;	// to check for arbitrary test term id already read

	//! this unsigned integer is one of [0,...,TOTAL_WORDS_IN_V_TEST]
	size_t _sequential_test_vocabulary_index;
	//! an entry in this std::vector corresponds to one in std::vector [0,...,TOTAL_WORDS_IN_V_TEST-1]
	std::vector<size_t> _test_datum_id_vec;
	/*@}*/

public:
	/**
	 * @brief returns a pointer to the SCorpus representing the training corpus
	 *
	 * @return a pointer to the SCorpus representing the training corpus
	 * @see SCorpus
	 */
	virtual inline SCorpus* get_training_corpus() { return _corpus; }
	/**
	 * @brief returns a pointer to the SCorpus representing the test or validation corpus
	 *
	 * @return a pointer to the SCorpus representing the test or validation corpus
	 * @see SCorpus
	 */
	virtual inline SCorpus* get_test_corpus() { return _corpus_test; }
	/**
	 * @brief returns a reference to the member variable _test_datum_id_to_sequential_index_id_map
	 * @see CLDAGeneralModel::_test_datum_id_to_sequential_index_id_map
	 *
	 * @return returns a reference to the member variable _test_datum_id_to_sequential_index_id_map which is an std::map<size_t, size_t>
	 * representing map<test_datum_id, sequential_id \\in [0,..., #unique_datum_id_in_test_set - 1]
	 */
	virtual inline std::map<size_t, size_t>& get_test_datum_id_to_sequential_index_id_map() { return _test_datum_id_to_sequential_index_id_map; }


public:
	/**
	 * @brief CLDAGibbs constructor
	 * This function takes care of the initializing all the member variables to their respective default values
	 */
	CLDAGibbs();
	/**
	 *
	 */
	~CLDAGibbs();

	/**
	 *
	 */
	virtual void set_bytecode(std::string bytecode);
	/**
	 *
	 */
	inline std::string get_bytecode() { return _bytecode; }
	/**
	 *
	 */
	virtual void set_model_dump_mode(std::string model_dump_mode) { CLDAGeneralModel::_model_dump_mode = model_dump_mode; }
	/**
	 *
	 */
	virtual void set_model_operation_mode(std::string model_operation_mode);
	/**
	 *
	 */
	virtual std::string get_model_operation_mode() { return CLDAGeneralModel::_model_operation_mode; }
	/**
	 *
	 */
	virtual std::string get_model_dump_mode() { return CLDAGeneralModel::_model_dump_mode; }
	/**
	 *
	 */
	virtual std::string get_model_root_dir() { return _model_root_dir; }
	/**
	 *
	 */
	virtual size_t get_test_vocab_size() { return _V_test; }
	/**
	 *
	 */
	virtual bool load_model();
	/**
	 *
	 */
	virtual void save_model();
	/**
	 *
	 */
	virtual void read_data_from_file();
	/**
	 *
	 */
	virtual void read_config_file(std::string model_config_filename);
	/**
	 *
	 */
	virtual inline std::string to_string() {
		if ( CLDAGeneralModel::_model_operation_mode == "train" ) return "LDA-Gibbs-train";
		if ( CLDAGeneralModel::_model_operation_mode == "test" || CLDAGeneralModel::_model_operation_mode == "validate" ) return "LDA-Gibbs-test";
		return "LDA-Gibbs";	// returns this std::string by default
	}
	/**
	 *
	 */
	virtual bool create_and_initialize_data_structures();
	/**
	 *
	 */
	virtual void reset_test_data_structures();
	/**
	 *
	 */
	virtual double compute_log_likelihood();
	/**
	 *
	 */
	virtual std::vector<std::vector<int>* >* get_train_doc_datum_topic_assignment_vectors() { return _z_d_n; }
	/**
	 *
	 */
	virtual std::vector<std::vector<int>* >* get_test_doc_datum_topic_assignment_vectors() { return _z_d_n_test; }
	/**
	 *
	 */
	virtual std::vector<std::vector<int>* >* get_train_doc_topic_assignment_vectors() { return _z_d_k; }
	/**
	 *
	 */
	virtual std::vector<std::vector<int>* >* get_test_doc_topic_assignment_vectors() { return _z_d_k_test; }
	/**
	 *
	 */
	virtual std::vector<std::vector<double>* >* get_train_topic_datum_multinomials() { return _beta; }
	/**
	 *
	 */
	virtual std::vector<std::vector<double>* >* get_train_doc_topic_proportion_multinomials() { return _theta; }
	/**
	 *
	 */
	virtual std::vector<std::vector<double>* >* get_test_doc_topic_proportion_multinomials() { return _theta_test; }
	/**
	 *
	 */
	virtual std::string get_datum_to_id_dictionary_filename() { return _datum_to_id_dict_filename; }
	/**
	 *
	 */
	virtual std::string get_train_saved_topic_multinomial_filename() { return _model_train_saved_topic_mult_filename; }
	/**
	 *
	 */
	virtual size_t get_train_doc_count() { return _D; }
	/**
	 *
	 */
	virtual size_t get_train_vocabulary_size() { return _V; }
	/**
	 *
	 */
	virtual long get_train_total_datum_count() { return _training_datum_total_count; }
	/**
	 * @brief computes the Bayesian Information Criterion from the model for the particular setting of model parameters
	 * @return the Bayesian Information Criterion from the model for the particular setting of model parameters
	 *
	 * By default returns 0 from base class
	 */
	virtual double get_BIC();
	/**
	 * @brief computes the Minimum Description Length from the model for the particular setting of model parameters
	 * @return the Minimum Description Length from the model for the particular setting of model parameters
	 *
	 * By default returns 0 from base class
	 */
	virtual double get_MDL();

protected:
	/**
	 *
	 */
	double compute_entropy(std::vector<double>* p);

	/**
	 *
	 */
	virtual void read_data_from_text_file();
	/**
	 *
	 */
	virtual void read_data_from_binary_file();
	/**
	 *
	 */
	virtual void read_grouped_data_from_text_file_for_training();
	/**
	 *
	 */
	virtual void read_grouped_data_from_text_file_for_testing();
	/**
	 *
	 */
	virtual void read_sequential_data_from_text_file_for_training();
	/**
	 *
	 */
	virtual void read_sequential_data_from_text_file_for_testing();
	/**
	 *
	 */
	virtual void read_grouped_data_from_binary_file_for_training();
	/**
	 *
	 */
	virtual void read_grouped_data_from_binary_file_for_testing();
	/**
	 *
	 */
	virtual void read_sequential_data_from_binary_file_for_training();
	/**
	 *
	 */
	virtual void read_sequential_data_from_binary_file_for_testing();
	/**
	 *
	 */
	virtual bool create_and_initialize_data_structures_for_training();
	/**
	 *
	 */
	virtual bool create_and_initialize_data_structures_for_testing();
	/**
	 *
	 */
	virtual void accommodate_for_unknown_datum_during_train_setup();
	/**
	 *
	 */
	virtual double compute_log_likelihood_for_training();
	/**
	 *
	 */
	virtual double compute_log_likelihood_for_testing();
	/**
	 *
	 */
	virtual bool load_model_in_text_mode();
	/**
	 *
	 */
	virtual void save_training_model_in_text_mode();
	/**
	 *
	 */
	virtual void save_training_model_in_binary_mode();
	/**
	 *
	 */
	virtual bool load_model_in_binary_mode();
	/**
	 *
	 */
	virtual void save_test_model_in_text_mode();
	/**
	 *
	 */
	virtual void save_test_model_in_binary_mode();

public:
	/**
	 *  some diagnostic functions
	 */
	virtual void print_parameters();

};

#endif /* CLDAGIBBS_H_ */
