/*
 * CModelGraph.h
 *
 *  Created on: May 19, 2010
 *      Author: pdas
 */

#ifndef CMODELGRAPH_H_
#define CMODELGRAPH_H_

#include "../include/includes.h"
#include "../include/CUtilities.h"
#include "AModelGraph.h"
//using namespace std;

struct SModelGraphNode;			// defined later in this file
class CModelGraphNodeCompare;	// defined later in this file
struct SForest;					// defined later in this file
struct SModelGraphSimpleNode;	// defined later in this file

/**
 * Main graphical model class
 */
class CModelGraph : public AModelGraph {

	std::map<SModelGraphNode*, std::list<SModelGraphNode*>*, CModelGraphNodeCompare> *_model_graph;
	std::set<SModelGraphNode*, CModelGraphNodeCompare>* _observed_variable_set;
	SForest* _model_forest_representation;

	// this map variable holds conjugate distribution std::strings like "Gam|DirSym" in lowercase
	// the key is the distribution for prior and the value is the distribution for the likelihood (parent->generated)
	std::multimap<std::string, std::string> _conjugate_distribution_multimap;

	// bytecode representation of the std::string
	std::string _model_bytecode;

public:
	CModelGraph();
	CModelGraph(CModelGraph& src_model_graph);
	CModelGraph& operator = (CModelGraph& src_model_graph );
	virtual ~CModelGraph();
	inline std::map<SModelGraphNode*, std::list<SModelGraphNode*>*, CModelGraphNodeCompare> * get_graph() { return _model_graph; }

	inline std::set<SModelGraphNode*, CModelGraphNodeCompare>* get_observed_variable_set() { return _observed_variable_set; }
	inline SForest* get_model_forest_representation() { return _model_forest_representation; }

	std::string to_string();
	void print_graph() const;

	void build_complete_graph();

	void reverse_graph();	// not implemented yet
	// this function returns a forest of maps of (level, distribution_name_list)
	void level_order_bfs(SForest* model_forest_representation);

	/**
	 * ///////////////////////////////////////////////////////////////////////////////////////////
	 * // based on the forest decide the model													//
	 * // essentially the model is a functionoid that has a couple of data initializer routines //
	 * // and a couple of sampling routines.													//
	 * // the main objective of this function is to decide how to initialize data matrices and  //
	 * // how then to call which sampling functions corresponding to variable conjugacy			//
	 *
	 */
	std::string match_compiled_template(std::string compiled_model_filename);
	/*
	 * return the model bytecode as a std::string
	 */
	inline std::string get_bytecode() { return _model_bytecode; }
	/*
	 * this function generates a std::vector of all possible paths from observed variables to priors/hyperpriors
	 * @param the model bytecode as a std::string
	 * @return std::list< std::pair<SModelGraphSimpleNode, std::list<std::list<SModelGraphSimpleNode> > > > -
	 * 												Each observed variable can have a list of paths and each path is a list of nodes
	 * 												Note: a copy is returned since the graph is small and we don't need to bother about
	 * 													destructors. An alternative will be to create a structure SPaths
	 */
	std::list< std::pair<SModelGraphSimpleNode, std::list<std::list<SModelGraphSimpleNode> > > > get_all_paths_from_all_obs_vars(std::string model_bytecode);

private:
	/**
	 *
	 */
	void add_to_conjugate_distribution_multimap();
	/**
	 *
	 */
	std::map<int, std::set<std::string>* >* level_order_bfs_subgraph(
									std::map<SModelGraphNode*, std::list<SModelGraphNode*>*, CModelGraphNodeCompare>* model_graph,
									SModelGraphNode* start);
	/**
	 * This function returns a deep copy of the source graph model and stores the pointer to the newly
	 * created graph in the instance variable _model_graph
	 */
	void deeply_copy_graphical_model(CModelGraph* src_graphical_model);
	/**
	 * auxilliary function to add numbers to a list thus appending the std::string with something like "(xx.yy,xxx.yyy,...)"
	 */
	void add_numeric_parameters_to_generating_random_variable(std::list<double>& l, std::string& s);
	/**
	 * auxilliary function to add numbers to a list thus appending the std::string with something like "(xx.yy,xxx.yyy,...)"
	 */
	void add_nonnumeric_parameters_to_generating_random_variable(std::list<std::string>& l, std::string& s);
	/**
	 * returns a new std::string that is a replica of s except with expressions like "(xx.yy,xxx.yyy,...)" deleted
	 */
	std::string delete_paranthesized_expression_from_bytecode_string(const std::string& s);

	/*
	 * function object for level set comparsions
	 */
	struct SStringSetCompareRemoveParenthesis {
		bool operator () (const std::string& s1, const std::string& s2) const {
			return delete_paranthesized_expression(s1) < delete_paranthesized_expression(s2) ;
		}
	private:
		std::string delete_paranthesized_expression(const std::string& s) const {
			std::string new_s = "";
			size_t length = s.length();
			bool paren_found = false;
			for ( size_t i = 0; i < length; ++i ) {
				if ( s.at(i) == '(' ) {
					paren_found = true;
				}
				if ( s.at(i) == ')' && paren_found ) {
					paren_found = false;
					continue;
				}
				if ( !paren_found )
					new_s += s.at(i);
			}
			return new_s;

		}
	};

	/**
	 * auxilliary functions to aid in finding paths from bytecode
	 */
	void find_paths( std::map<SModelGraphSimpleNode, std::list<SModelGraphSimpleNode> >& model_simple_graph,
			SModelGraphSimpleNode& node, std::list<std::list<SModelGraphSimpleNode> >& paths );
	void build_path_recursively( std::map<SModelGraphSimpleNode, std::list<SModelGraphSimpleNode> >& model_simple_graph,
			SModelGraphSimpleNode& node, std::list<SModelGraphSimpleNode>& path);
	/**
	 *
	 */
	template<class T>
	bool is_node_in_list(std::list<T>& adj_list, T& node) {
		typename std::list<T>::iterator l_itr;
		for ( l_itr = adj_list.begin(); l_itr != adj_list.end(); ++l_itr )
			if ( (*l_itr) == node )
				return true;
		return false;
	}

};

/**
 * The SModelGraphNode structure is the node that makes up the model graph. Each node contains:
 * -- name of the distribution (represented as std::std::string)
 * -- parameters of this distribution (represented as a std::vector of doubles)
 */
struct SModelGraphNode {
	// by default the members are public
	std::string _variable_name;				// name of this hidden or observed variable or parameter
	std::string _generating_distribution;	// name of this the generating distribution

	std::string _variable_indices;			// "plate" index of this variable (stored as a std::string of characters)
	std::string _dimensionality;				// dimensionality of this variable (stored as a std::string) - useful for checking matrix dimensions
	bool _observed;						// whether this variable is hidden or observed

	std::list<double> _generating_distribution_numeric_parameters_list;	// parameters for the hidden variable
	std::list<std::string> _generating_distribution_string_parameters_list;	// name of the hidden parent variable or parameter

	SModelGraphNode() {	}

	SModelGraphNode(std::string variable_name, std::string generating_distribution) {
		_variable_indices = "";
		_dimensionality = "";
		_variable_name = variable_name;
		_observed = false;
		_generating_distribution = generating_distribution;
	}
	~SModelGraphNode() {}

	// no need to define a copy constructor - a shallow copy is good enough for this node to be copied

	inline bool is_observed() { return _observed; }

	inline void set_variable_name(std::string variable_name) { _variable_name = variable_name; }
	inline void set_variable_indices(std::string indices) { _variable_indices = indices; }
	inline void set_dimensionality(std::string dim) { _dimensionality += dim; }
	inline void set_observed_flag(bool observed) { _observed = observed; }
	inline void set_generating_distribution_numeric_parameter_list(std::list<double> v) {
		_generating_distribution_numeric_parameters_list = v;	// assignment copy of list v
	}
	inline void set_generating_distribution_string_parameter_list(std::list<std::string> v) {
		_generating_distribution_string_parameters_list = v;	// assignment copy of list v
	}

	inline std::string to_string() {
		std::list<std::string>::iterator list_string_itr;
		std::list<double>::iterator list_double_itr;

		std::string s = "[ " + _variable_name + "," + _variable_indices + "," + _generating_distribution;
			s += "[";
				for ( list_double_itr = _generating_distribution_numeric_parameters_list.begin();
					list_double_itr != _generating_distribution_numeric_parameters_list.end(); list_double_itr++ )
					s += CUtilities::number_to_string<double>(*list_double_itr) + ",";
			s += "]";
			s += "[";
				for ( list_string_itr = _generating_distribution_string_parameters_list.begin();
					list_string_itr != _generating_distribution_string_parameters_list.end(); list_string_itr++ )
					s += (*list_string_itr) + ",";
			s += "]";
		s += "," + _dimensionality	+ ", @observed:" + (_observed ? "true" : "false") + " ]";
		return s;
	}


};

/**
 * The function object for graph node compare
 */
class CModelGraphNodeCompare { // simple comparison function based on distribution name
   public:
      bool operator()(const SModelGraphNode *x, const SModelGraphNode *y) const {
    	  return x->_variable_name < y->_variable_name;
      } // returns x<y
};

/**
 * this structure is for internal use to bfs
 */
struct SNodeInfo {
	int _depth;
	SModelGraphNode* _predecessor;
	int _visit_status;
	int _discovery_time; 	// only for DFS
	int _finish_time; 		// only for DFS
	SNodeInfo() {
		_finish_time = 0;
		_discovery_time = 0;
	}
	inline std::string to_string() {
		std::string s = "[ ";
			s += CUtilities::number_to_string<int>(_depth) + ", "+_predecessor->to_string()+"; ";
		s += "] ";
		return s;
	}
};

/**
 * The observed variable forest - useful for comparing different graphical models to some compiled ones
 * essetially a std::map<std::string,std::map<int, std::set<std::string>*>*>* with a structure breakdown like :
 * [std::map<observed_variable_name, m>]
 * w (observed) >
 * [m = std::map<    int, {std::string1,.., std::stringN} >]
 *		[level] 0 : mult,
 *		[level] 1 : dirsym, mult,
 *		[level] 2 : dirsym,
 *
 *	Each observed variable should form the root of a tree in the forest
 *
 */
struct SForest {
	std::map<std::string,std::map<int, std::set<std::string>*>*>* _forest;
	SForest() {
		_forest = new std::map<std::string,std::map<int, std::set<std::string>*>*>();
	}
	~SForest() {
		std::map<std::string,std::map<int, std::set<std::string>*>*>::iterator f_itr;
		std::map<int, std::set<std::string>*>::iterator t_itr;
		for ( f_itr = _forest->begin(); f_itr != _forest->end(); f_itr++ ) {
			std::map<int, std::set<std::string>*>* tree = f_itr->second;
			for ( t_itr = tree->begin(); t_itr != tree->end(); t_itr++ ) {
				delete t_itr->second;
			}
			delete tree;
		}
		delete _forest;
	}

	std::string to_string() {
		std::string s = "[\n";
		std::map<std::string, std::map<int, std::set<std::string>*>*>::iterator f_itr;
		std::map<int, std::set<std::string>*>::iterator t_itr;
		std::set<std::string>::iterator s_itr;
		int level;

		for (f_itr = _forest->begin(); f_itr != _forest->end(); f_itr++) {
			s += f_itr->first + ">\n";
			std::map<int, std::set<std::string>*>* tree = f_itr->second;
			for (t_itr = tree->begin(); t_itr != tree->end(); t_itr++) {
				level = t_itr->first;
				s += "\tlevel " + CUtilities::number_to_string<int>(level) + " : ";
				for ( s_itr = t_itr->second->begin(); s_itr != t_itr->second->end(); s_itr++ )
					s += *s_itr + ", ";
				s += "\n";
			}
		}
		s += "]\n";
		return s;
	}

};

/**
 * The SModelGraphNode structure is the node that makes up the model graph. Each node contains:
 * -- name of the distribution (represented as std::std::string)
 * -- parameters of this distribution (represented as a std::vector of doubles)
 */
struct SModelGraphSimpleNode {
	// by default the members are public
	std::string _variable_name;				// name of this hidden or observed variable or parameter
	std::string _dimensionality;				// dimensionality of this variable (stored as a std::string) - useful for checking matrix dimensions
	bool _observed;						// whether this variable is hidden or observed
	std::list<std::string> _parameter_list;		// name of the hidden parent variable or parameter

	SModelGraphSimpleNode() {	}

	SModelGraphSimpleNode(std::string variable_name) {
		_dimensionality = "";
		_variable_name = variable_name;
		_observed = false;
	}
	~SModelGraphSimpleNode() {}

	// no need to define a copy constructor - a shallow copy is good enough for this node to be copied

	inline bool is_observed() { return _observed; }

	inline void set_variable_name(std::string variable_name) { _variable_name = variable_name; }
	inline void set_dimensionality(std::string dim) { _dimensionality += dim; }
	inline void set_observed_flag(bool observed) { _observed = observed; }
	inline void set_string_parameter_list(std::list<std::string> v) {
		_parameter_list = v;	// assignment copy of list v
	}

	inline std::string to_string() {
		std::list<std::string>::iterator list_string_itr;

		std::string s = "[ " + _variable_name + ", @obs?" + (_observed ? "true" : "false") + " " +  _dimensionality;
			s += " (";
				for ( list_string_itr = _parameter_list.begin();
					  list_string_itr != _parameter_list.end(); ++list_string_itr )
					s += (*list_string_itr) + ", ";
			s += ")";
		s += " ]";
		return s;
	}

	bool operator < (const SModelGraphSimpleNode& y) const {
		return _variable_name < y._variable_name;
	} // returns x<y

	bool operator == (const SModelGraphSimpleNode& y) const {
		return _variable_name == y._variable_name;
	} // returns x == y

};


#endif /* CMODELGRAPH_H_ */
