/*
 * TemplateMath.h
 *
 *  Created on: Mar 1, 2010
 *      Author: pdas3
 */

#ifndef TEMPLATEMATH_H_
#define TEMPLATEMATH_H_

template<typename T>
class CTemplateMatrixMath {

public:

	static T max_three_numbers(const T& a, const T& b, const T& c);
	static void max_value_indices_in_array(T *v, size_t num_elements, int *index);

	// templatized matrix manipulation functions
	static void zero_initialize_matrix(std::vector<std::vector<T>* >* mat);
	static void zero_initialize_vector(std::vector<T>* vec);
	static void create_2D_matrix(std::vector<std::vector<T>* >** mat, size_t n_rows, size_t n_cols);
	static void delete_2D_matrix(std::vector<std::vector<T>* >* mat);

};

template<typename T>
T CTemplateMatrixMath<T>::max_three_numbers(const T& a, const T& b, const T& c) {
	T maxi = a;
	if ( b > a ) maxi = b;
	if ( c > maxi ) maxi = c;
	return maxi;
}

template<typename T>
void CTemplateMatrixMath<T>::max_value_indices_in_array(T* v, size_t num_elements, int *index) {
	T mx = std::numeric_limits<T>::min();
	for (size_t i = 0; i < num_elements; ++i)
		if ( v[i] > mx ) mx = v[i];

	for (size_t i = 0; i < num_elements; ++i) {
		index[i] = 0;
		if ( v[i] == mx ) index[i] = 1;
	}
	return;
}

/**
 * some common memory allocation routines for matrices represented as std::vectors
 */
template <typename T>
void CTemplateMatrixMath<T>::zero_initialize_matrix(std::vector<std::vector<T>* >* mat)
{
	for ( size_t i = 0; i < mat->size(); i++ )
		for ( size_t j = 0; j < mat->at(i)->size(); j++ )
			mat->at(i)->at(j) = 0;
}
template <typename T>
void CTemplateMatrixMath<T>::zero_initialize_vector(std::vector<T>* vec)
{
	for ( size_t i = 0; i < vec->size(); i++ )
		vec->at(i) = 0;
}
template <typename T>
void CTemplateMatrixMath<T>::create_2D_matrix(std::vector<std::vector<T>* >** mat, size_t n_rows, size_t n_cols)
{
	//if (b_verbose_global) std::cout << nRows << "x" << nCols << std::endl;
	*mat = new std::vector< std::vector<T>* >(n_rows);
	for ( size_t i = 0 ; i < n_rows ; i++ ) {
		std::vector<T>* row = new std::vector<T>(n_cols, 0);
		(*mat)->at(i) = row;
	}
}
template <typename T>
void CTemplateMatrixMath<T>::delete_2D_matrix(std::vector<std::vector<T>* >* mat)
{
	size_t n_rows = mat->size();
	for ( size_t i = 0 ; i < n_rows ; i++ ) {
		delete mat->at(i);
	}
	delete mat;
	mat = 0x0;
}

/**
 *
 */
template<typename T>
class CMatrix2D {
	/**
	 * The 2D matrix is a std::vector of std::vectors
	 */
	std::vector<std::vector<T>* >* _mat;
	size_t _rows;
	size_t _cols;

public:

	CMatrix2D () {
		_rows = 0;
		_cols = 0;
	}

	/**
	 * constructor
	 */
	CMatrix2D(size_t rows, size_t cols) {
		CTemplateMatrixMath<T>::create_2D_matrix(_mat, rows, cols);
		_rows = rows;
		_cols = cols;
	}

	/**
	 * destructor
	 */
	~CMatrix2D() {
		CTemplateMatrixMath<T>::delete_2D_matrix(_mat);
	}

	size_t get_rows() { return _rows; }
	size_t get_cols() { return _cols; }

	/**
	 * overloading operator () for matrix element access
	 */
	inline T& operator() (size_t r, size_t c) {
		if (r < 0 || r > _rows ) {
			if (b_verbose_global) std::cout<<"Illegal row index";
			return 0;
		} else if ( c < 0 || c > _cols) {
			if (b_verbose_global) std::cout<<"Illegal Column Index:";
			return 0;
		} else return (*(*_mat)[r])[c];
	}

	/**
	 * overloading << operator for stream insertion operations
	 */
	inline friend std::ostream& operator<< (std::ostream& os, CMatrix2D<T>& mat){
		size_t rows = mat.get_rows();
		size_t cols = mat.get_cols();
		for(size_t i = 0; i < rows; ++i) {
			for(size_t j=0; j < cols; ++j)
				os << mat(i,j) << " ";
			os << std::endl;
		}
		return os;
	}

	/**
	 * overloading = operator
	 */
	inline CMatrix2D<T>& operator= (CMatrix2D<T>& mat){
		size_t rows = mat.get_rows();
		size_t cols = mat.get_cols();
		CTemplateMatrixMath<T>::create_2D_matrix(this->_mat, rows, cols);
		for (size_t i = 0; i < rows; ++i )
			for(size_t j = 0; j < cols; ++j )
				this->_mat(i,j) = mat(i,j);

		return *this;
	}

	/**
	 * copy constructor
	 */
	CMatrix2D(CMatrix2D<T>& mat){
		size_t rows = mat.get_rows();
		size_t cols = mat.get_cols();
		CTemplateMatrixMath<T>::create_2D_matrix(this->_mat, rows, cols);
		for (size_t i = 0; i < rows; ++i )
			for(size_t j = 0; j < cols; ++j )
				this->_mat(i,j) = mat(i,j);
		return;
	}

	/**
	 * reset elements in the matrix to 0
	 */
	inline void reset() {
		CTemplateMatrixMath<T>::zero_initialize_matrix(this->_mat);
	}

	/**
	 * print the matrix
	 */
	inline void print() {
		for (size_t i = 0; i < _rows; ++i ) {
			for(size_t j = 0; j < _cols; ++j )
				if (b_verbose_global) std::cout << _mat(i,j) << " ";
			if (b_verbose_global) std::cout << std::endl;
		}
	}
};

#endif /* TEMPLATEMATH_H_ */
