/**
 * @mainpage Core methods for implementing topic models
 * @version 1.0
 * @author Pradipto Das
 * @date August 2010
 * @note coding Topic Models in an object oriented way

	This manual is divided in the following sections:
	- @subpage intro
	- @subpage advanced "Advanced usage"
 */


/** @page intro Introduction
This toolbox lays down a framework for topic modeling of unstructured text documents. The basic Latent Dirichlet Allocation (LDA) model
is implemented here, however, the code design is extensible enough to plug in any other model without much effort. This toolbox is a
slight deviation from Hal Daume's Hierarchical Bayesian Compiler (HBC) in that this toolbox does not generate code from a sample model
configuration file. Rather, it tries to recognize what model does a sample model configuration file represents and calls the handler
to execute such a model. In one view, this toolbox is trying to act as an interpreter for graphical models so that the individual model
codes are both human readable and "tweakable" for different kinds of optimizations.

The aim of this project is to have a stream lined robust API for topic model prediction tasks.

You can now proceed to the @ref advanced "advanced section".
 */


/** @page advanced Advanced Usage
This page is for advanced users.
Make sure you have first read @ref intro "the introduction".
This code is an implementation of the paper:

"Finding scientific topics," Griffiths, Thomas L. and Steyvers, Mark, "Proceedings of the National Academy of Sciences of the United States of America," 2004

For the original implementation of Latent Dirichlet Allocation using Variational Bayes, see David Blei's original paper titled "Latent Dirichlet Allocation"
published in Journal of Machine Learning Research.

@see http://www.cs.princeton.edu/~blei/publications.html
 */



#include "../include/includes.h"

#include "CModelTestCPPUnit.h"
#include "CModelValidateCPPUnit.h"
#include "CModelTrainCPPUnit.h"

using namespace std;

bool b_verbose_global;

/**
 * @brief This file TopicModelToolbox-Sampler.cpp contains the main() method that illustrates some of the ways
 * in which the topic model toolbox can be used. This toolbox is in the initial phases of development and should
 * undergo extensive additions and modifications by me/later developers
 *
 * As of December 01, 2010, the toolbox implements the Gibbs sampling version of Latent Dirichlet Allocation (LDA)
 * model that extracts themes from documents with each theme being a distribution over a set of words that belong to
 * the corpus vocabulary
 *
 * Note that the LDA model is restricted to discrete data but not restricted to text only
 *
 * @param argc is the number of arguments to the executable
 * @param argv is the array of character arrays that hold the argument values supplied from command line or so
 *
 *  arg[0] = [string] program name \n
 *  arg[1] = [string] model_description_file \n
	arg[2] = [string] model_compiled_file \n
	arg[3] = [string] model_config_file \n
	arg[4] = [string] mode \n
	arg[5] = [string] [--verbose]\n

	@note a sample training model_description_file reads:

	@verbatim
input file: _sample_data/YahooAnswersSmall09/index/topicmodel.sequential.input
input dict file: _sample_data/YahooAnswersSmall09/index/term_to_id.txt
data input mode: sequential
num topics: 10
gibbs iter: 2000
model root dir: _sample_data/YahooAnswersSmall09/gibbs_lda_store/train/10
gibbs iteration lag: 200
gibbs iteration burnin: 1000
	@endverbatim

	@note a sample test model_description_file reads:

	@verbatim
input file: _sample_data/index/YahooAnswersSmall09/index/topicmodel.sequential.input
data input mode: sequential
input dict file: _sample_data/index/YahooAnswersSmall09/index/term_to_id.txt
num topics: 10
gibbs iter: 1500
model root dir: _sample_data/index/YahooAnswersSmall09/gibbs_lda_store/test/10
gibbs iteration lag: 50
gibbs iteration burnin: 1000
model train root dir: _sample_data/index/YahooAnswersSmall09/gibbs_lda_store/train/10
	@endverbatim

	@note a sample cross-validation model_description_file reads:

	@verbatim
input file: _sample_data/index/YahooAnswersSmall09/index/topicmodel.sequential.input
data input mode: sequential
input dict file: _sample_data/index/YahooAnswersSmall09/index/term_to_id.txt
topic range: 2 8 12 16 20
model root dir: _sample_data/index/YahooAnswersSmall09/gibbs_lda_store/cross-val
gibbs iter: 2000
gibbs iteration lag: 100
gibbs iteration burnin: 1000
gibbs iter validate: 1000
gibbs iteration lag validate: 50
gibbs iteration burnin validate: 500
num folds: 10
	@endverbatim
	OR\n
	@verbatim
input file: _sample_data/index/YahooAnswersSmall09/index/topicmodel.input
data input mode: grouped
input dict file: _sample_data/index/YahooAnswersSmall09/index/term_to_id.txt
topic range: 2 8 12 16 20
model root dir: _sample_data/index/YahooAnswersSmall09/gibbs_lda_store/cross-val
gibbs iter: 2000
gibbs iteration lag: 100
gibbs iteration burnin: 1000
gibbs iter validate: 1000
gibbs iteration lag validate: 50
gibbs iteration burnin validate: 500
num folds: 10
	@endverbatim

	@note a sample model_compiled_file reads:

	@verbatim
LDA-dirsym=mult$dirsym|mult$dirsym|mult,gam|dirsym$gam|dirsym$
LDA-dirsym=mult$dirsym|mult$dirsym|mult$
LDA-dirasym=mult$dir|mult$dir|mult$
	@endverbatim

	@note a sample model_config file reads:

	@verbatim
beta_{k}	~ DirSym(0.01, V)
theta_{d}	~ DirSym(0.01, K)
z_{d,n}		~ Mult(theta_{d})
w_{d,n}		~ mult(beta_{z_{d,n}})	@observed
	@endverbatim

	@note mode can be "train", "test" or "cross-validate"
 *
 */
int main(int argc, char** argv) {

	b_verbose_global = false;

	string verbose;
	string model_description_file;
	string model_compiled_file;
	string model_config_file;
	string mode;

	if ( argc < 5 ) {
		CUtilities::print_command_help();
		return 0;
	} else if ( argc >= 5 ) {
		model_description_file = argv[1];
		model_compiled_file = argv[2];
		model_config_file = argv[3];
		mode = argv[4];
	}
	if ( argc == 6 ) verbose = argv[5];
	if ( verbose == "--verbose" ) b_verbose_global = true;

	//////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// Let the appropriate model take over now
	//////////////////////////////////////////////////////////////////////////////
	if ( mode == "train" )
		CModelTrainCPPUnit unit_test_train_model(model_description_file, model_compiled_file, model_config_file);

	if ( mode == "cross-validate" ) {

		CModelValidateCPPUnit unit_test_model_validate(model_description_file, model_compiled_file, model_config_file);

	}

	if ( mode == "test" )
		CModelTestCPPUnit unit_test_model_test(model_description_file, model_compiled_file, model_config_file);

	return 0;
} // main ends here

